import json

import requests_mock

from jf_ingest.config import IngestionConfig, IngestionType
from jf_ingest.file_operations import IngestIOHelper, SubDirectory


class TestIngestIOHelper:
    company_slug = "test_company"
    timestamp = "20231205_123456"
    local_file_path = "/tmp"
    api_token = "test_token"
    ingest_io_helper = None

    def _get_ingest_io_helper(self) -> IngestIOHelper:
        return IngestIOHelper(ingest_config=self._get_ingest_config())

    def _get_ingest_config(self):
        return IngestionConfig(
            company_slug=self.company_slug,
            timestamp=self.timestamp,
            local_file_path=self.local_file_path,
            jellyfish_api_token=self.api_token,
        )

    def setup(self):
        self.ingest_io_helper = self._get_ingest_io_helper()

    def test_write_and_read_file(self):
        self.ingest_io_helper.write_json_data_to_local(
            "jira_test_data", {"test": "data"}, SubDirectory.JIRA
        )
        with open("/tmp/jira/jira_test_data.json", "r") as f:
            data = json.load(f)
            assert data == {"test": "data"}

    def test_upload_to_s3_with_mocked_s3_direct_connect(self):
        filepath = "jira_test_data.json"
        s3_obj_path = f"{self.company_slug}/{self.timestamp}/jira/jira_test_data.json"
        s3_url = "https://jellyfish-agent-upload.s3.amazonaws.com/"
        json_resp = {
            "signed_urls": [
                {
                    "url": f"{s3_url}",
                    "fields": {
                        "key": f"{self.company_slug}/{self.timestamp}/jira/{filepath}.gz",
                        "AWSAccessKeyId": "PLACEHOLDER",
                        "x-amz-security-token": "PLACEHOLDER",
                    },
                }
            ]
        }

        with requests_mock.Mocker() as m:
            m: requests_mock.Mocker = m
            m.register_uri(
                method="POST",
                url="https://app.jellyfish.co/endpoints/ingest/signed-url?timestamp=20231205_123456",
                json=json_resp,
            )
            m.register_uri(method="POST", url=s3_url, status_code=200)

            signed_url = self.ingest_io_helper.get_signed_urls(
                [filepath], self.timestamp, ingestion_type=IngestionType.DIRECT_CONNECT
            )[0]

            assert m.last_request.json()["ingestType"] == "DIRECT_CONNECT"
            self.ingest_io_helper.write_json_data_to_local(
                filepath, {"test": "data"}, SubDirectory.JIRA
            )
            success = self.ingest_io_helper.wrapped_upload_file_to_s3(
                f"{self.local_file_path}/{SubDirectory.JIRA}/{filepath}",
                s3_obj_path,
                signed_url,
            )
            assert success

    def test_upload_to_s3_with_mocked_s3_agent(self):
        filepath = "jira_test_data.json"
        s3_obj_path = f"{self.company_slug}/{self.timestamp}/jira/jira_test_data.json"
        s3_url = "https://jellyfish-agent-upload.s3.amazonaws.com/"
        json_resp = {
            "signed_urls": [
                {
                    "url": f"{s3_url}",
                    "fields": {
                        "key": f"{self.company_slug}/{self.timestamp}/jira/{filepath}.gz",
                        "AWSAccessKeyId": "PLACEHOLDER",
                        "x-amz-security-token": "PLACEHOLDER",
                    },
                }
            ]
        }

        with requests_mock.Mocker() as m:
            m: requests_mock.Mocker = m
            m.register_uri(
                method="POST",
                url="https://app.jellyfish.co/endpoints/ingest/signed-url?timestamp=20231205_123456",
                json=json_resp,
            )
            m.register_uri(method="POST", url=s3_url, status_code=200)

            signed_url = self.ingest_io_helper.get_signed_urls(
                [filepath], self.timestamp, ingestion_type=IngestionType.AGENT
            )[0]

            assert m.last_request.json()["ingestType"] == "AGENT"

            self.ingest_io_helper.write_json_data_to_local(
                filepath, {"test": "data"}, SubDirectory.JIRA
            )
            success = self.ingest_io_helper.wrapped_upload_file_to_s3(
                f"{self.local_file_path}/{SubDirectory.JIRA}/{filepath}",
                s3_obj_path,
                signed_url,
            )
            assert success



import logging
import os
import sys
from datetime import datetime

import urllib3
import yaml

from jf_ingest.file_operations import IngestIOHelper
from jf_ingest.jf_jira import IngestionConfig, load_and_push_jira_to_s3
from jf_ingest.jf_jira.auth import JiraConfig


def setup_harness_logging(logging_level: int):
    """Helper function to setting up logging in the harness"""
    logging.basicConfig(
        level=logging_level,
        format="%(asctime)s %(threadName)s %(levelname)s %(name)s %(message)s"
        if logging_level == logging.DEBUG
        else "%(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
    )
    logging.getLogger(urllib3.__name__).setLevel(logging.WARNING)


if __name__ == "__main__":
    """
    NOTE: This is a work in progress developer debugging tool.
    it is currently run by using the following command:
       pdm run ingest_harness [--debug]
    and it requires you to have a creds.env and a config.yml file at
    the root of this project
    """
    debug_mode = "--debug" in sys.argv

    # Get Config data for Ingestion Config
    with open("./config.yml") as yaml_file:
        yaml_data = yaml.safe_load(yaml_file)
        general_config_data = yaml_data["general"]
        jira_config_data = general_config_data["jira_config"]
        jira_config_data["work_logs_pull_from"] = datetime.strptime(
            jira_config_data.get("earliest_issue_dt", datetime.min), "%Y-%m-%d"
        )
        jira_config_data["earliest_issue_dt"] = datetime.strptime(
            jira_config_data.get("earliest_issue_dt", datetime.min), "%Y-%m-%d"
        )
        ingest_config = IngestionConfig(
            timestamp=datetime.utcnow().strftime("%Y%m%d_%H%M%S"),
            jellyfish_api_token=os.getenv("JELLYFISH_API_TOKEN"),
            **general_config_data,
        )
        ingest_config.company_slug = (os.getenv("COMPANY_SLUG"),)
        jira_config = JiraConfig(**jira_config_data)
        jira_config.url = os.getenv("JIRA_URL")
        jira_config.user = os.getenv("JIRA_USERNAME")
        jira_config.password = os.getenv("JIRA_PASSWORD")
        ingest_config.local_file_path = f"{ingest_config.local_file_path}/{ingest_config.timestamp}"
        # Inject auth data into config
        ingest_config.jira_config = jira_config

    setup_harness_logging(logging_level=logging.DEBUG if debug_mode else logging.INFO)

    load_and_push_jira_to_s3(ingest_config)



import enum
import tempfile
from dataclasses import dataclass, field
from datetime import datetime

from dateutil import parser

from jf_ingest.utils import format_datetime_to_ingest_timestamp

JELLYFISH_API_BASE = "https://app.jellyfish.co"


class JiraAuthMethod(enum.Enum):
    BasicAuth = 1
    AtlassianConnect = 2


# Needs to subclass str to be serializable
class IngestionType(str, enum.Enum):
    AGENT = "AGENT"
    DIRECT_CONNECT = "DIRECT_CONNECT"


@dataclass
class IssueMetadata:
    id: str
    key: str
    updated: datetime
    project_id: str = (
        None  # NOTE: This field is optionally set, and generally only used for detected re-keys
    )
    # The following fields are used for detecting redownloads
    epic_link_field_issue_key: str = None
    parent_field_issue_key: str = None
    parent_id: str = None

    def __post_init__(self):
        """Post Init is called here to properly type everything by doing typecasts"""
        self.id = str(self.id)
        self.key = str(self.key)
        self.updated = (
            self.updated if isinstance(self.updated, datetime) else parser.parse(self.updated)
        )

        # Sanity recasts to make sure everything is a string
        if self.project_id:
            self.project_id = str(self.project_id)
        if self.epic_link_field_issue_key:
            self.epic_link_field_issue_key = str(self.epic_link_field_issue_key)
        if self.parent_field_issue_key:
            self.parent_field_issue_key = str(self.parent_field_issue_key)
        if self.parent_id:
            self.parent_id = str(self.parent_id)

    @staticmethod
    def init_from_jira_issue(issue: dict, project_id: str = None, skip_parent_data: bool = False):
        fields: dict = issue.get("fields", {})
        return IssueMetadata(
            id=issue["id"],
            key=issue["key"],
            project_id=project_id,
            updated=parser.parse(fields.get("updated")) if fields.get("updated") else None,
            parent_id=fields.get("parent", {}).get("id") if not skip_parent_data else None,
            parent_field_issue_key=fields.get("parent", {}).get("key")
            if not skip_parent_data
            else None,
        )

    @staticmethod
    def init_from_jira_issues(issues=list[dict], skip_parent_data: bool = False):
        return [
            IssueMetadata.init_from_jira_issue(issue, skip_parent_data=skip_parent_data)
            for issue in issues
        ]

    # Define a hashing function so that we can find uniqueness
    # easily using sets
    def __hash__(self) -> str:
        return hash(self.id)

    def __eq__(self, __o) -> bool:
        return hash(self) == hash(__o)


@dataclass
class JiraConfig:
    company_slug: str = None
    url: str = None
    # NOTE: Used in the User-Agent header. Not related
    # to the Jellyfish Agent
    user_agent: str = "jellyfish/1.0"
    # Used for Basic Auth
    user: str = None
    password: str = None
    personal_access_token: str = None
    # Used for Atlassian Direct Connect
    jwt_attributes: dict[str, str] = field(default_factory=dict)
    bypass_ssl_verification: bool = False
    required_email_domains: bool = False
    is_email_required: bool = False
    available_auth_methods: list[JiraAuthMethod] = field(default_factory=list)
    connect_app_active: bool = False
    # Jira Server Information
    gdpr_active: bool = None

    include_fields: list[str] = None
    exclude_fields: list[str] = None
    # User information
    force_search_users_by_letter: bool = False
    search_users_by_letter_email_domain: str = None

    # Projects information
    include_projects: list[str] = None
    exclude_projects: list[str] = None
    include_project_categories: list[str] = None
    exclude_project_categories: list[str] = None

    # Boards/Sprints
    download_sprints: bool = False

    # Issues
    skip_issues: bool = False
    only_issues: bool = False
    full_redownload: bool = False
    earliest_issue_dt: datetime = datetime.min
    issue_download_concurrent_threads: int = 1
    # Dict of Issue ID (str) to IssueMetadata Object
    jellyfish_issue_metadata: list[IssueMetadata] = None
    jellyfish_project_ids_to_keys: dict = None

    # worklogs
    download_worklogs: bool = False
    # Potentially solidify this with the issues date, or pull from
    work_logs_pull_from: datetime = datetime.min

    # Jira Ingest Feature Flags
    feature_flags: dict = field(default_factory=dict)


@dataclass
class IngestionConfig:
    # upload info
    upload_to_s3: bool = False
    local_file_path: str = None
    company_slug: str = None
    timestamp: str = None

    # Jira Auth Info
    jira_config: JiraConfig = None

    # JF specific config
    jellyfish_api_token: str = None
    jellyfish_api_base: str = JELLYFISH_API_BASE
    ingest_type: IngestionType = None

    def __post_init__(self):
        if not self.timestamp:
            self.timestamp = format_datetime_to_ingest_timestamp(datetime.utcnow())

        if not self.local_file_path:
            self.local_file_path = f"{tempfile.TemporaryDirectory().name}/{self.timestamp}"



import logging
import os
from contextlib import contextmanager
from functools import wraps

logger = logging.getLogger(__name__)


LOG_FILE_NAME = "jf_agent.log"


def log_entry_exit(*args, **kwargs):
    def actual_decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            func_name = func.__name__
            logger.debug(f"{func_name}: Starting")
            ret = func(*args, **kwargs)
            logger.debug(f"{func_name}: Ending")
            return ret

        return wrapper

    return actual_decorator


@contextmanager
def log_loop_iters(loop_desc, this_iternum, log_every_n_iters, log_entry=True, log_exit=True):
    if (this_iternum - 1) % log_every_n_iters == 0:
        if log_entry:
            logger.debug(f'Loop "{loop_desc}", iter {this_iternum}: Starting')
        yield
        if log_exit:
            logger.debug(f'Loop "{loop_desc}", iter {this_iternum}: Ending')
    else:
        yield


# Mapping of error/warning codes to templated error messages to be called by
# log_standard_error(). This allows for Jellyfish to better categorize errors/warnings.
ERROR_MESSAGES = {
    0000: "An unknown error has occurred. Error message: {}",
    3000: "Failed to upload file {} to S3 bucket",
    3001: "Connection to bucket was disconnected while uploading: {} with the following error: {}. Retrying...",
    3010: "Rate limiter: thought we were operating within our limit (made {}/{} calls for {}), but got HTTP 429 anyway!",
    3020: "Next available time to make call is after the timeout of {} seconds. Giving up.",
    3030: "ERROR: Could not parse response with status code {}. Contact an administrator for help.",
    3011: "Error normalizing PR {} from repo {}. Skipping...",
    3021: "Error getting PRs for repo {}. Skipping...",
    3031: "Unable to parse the diff For PR {} in repo {}; proceeding as though no files were changed.",
    3041: "For PR {} in repo {}, caught HTTPError (HTTP 401) when attempting to retrieve changes; "
    "proceeding as though no files were changed",
    3051: "For PR {} in repo {}, caught UnicodeDecodeError when attempting to decode changes; proceeding as though no files were changed",
    3061: "Failed to download {} data:\n{}",
    3071: "Jira rate limit exceeded on func {}, retry {} / {}.  Trying again in {}...",
    3081: "Got unexpected HTTP 403 for repo {}.  Skipping...",
    3091: "Github rate limit exceeded.  Trying again in {}...",
    3101: "Request to {} has failed {} times -- giving up!",
    3121: 'Got HTTP {} when fetching commit {} for "{}", this likely means you are trying to fetch an invalid re',
    3131: "Got {} {} when {} ({})",
    3141: "Got {} {} when {}",
    3151: "Getting HTTP 429s for over an hour; giving up!",
    3002: "Failed to download jira data:\n{}",
    3012: "Caught KeyError from search_issues(), reducing batch size to {}",
    3022: "Caught KeyError from search_issues(), batch size is already 0, bailing out",
    3032: "Exception encountered in thread {}\n{}",
    3042: "[Thread {}] Jira issue downloader FAILED",
    3052: "JIRAError ({}), reducing batch size to {}",
    3062: "Apparently unable to fetch issue based on search_params {}",
    3072: "Error calling createmeta JIRA endpoint",
    3082: "OJ-9084: Changelog history item with no 'fieldId' or 'field' key: {}",
    3092: (
        "OJ-22511: server side 500, batch size reduced to 0, "
        "last error was: {} with jql: {}, start_at: {}, and batch_size: {}. Skipping one issue ahead..."
    ),
    2101: "Failed to connect to {}:\n{}",
    2102: "Unable to access project {}, may be a Jira misconfiguration. Skipping...",
    2112: "Failed to connect to Jira for project ID: {}",
    2122: "you do not have the required 'development field' permissions in jira required to scan for missing repos",
    2132: (
        "Missing recommended jira_fields! For the best possible experience, "
        "please add the following to `include_fields` in the "
        "configuration file: {}"
    ),
    2142: (
        "Excluding recommended jira_fields! For the best possible experience, "
        "please remove the following from `exclude_fields` in the "
        "configuration file: {}",
    ),
    2201: (
        "\nERROR: Failed to download ({}) repo(s) from the group {}. "
        "Please check that the appropriate permissions are set for the following repos... ({})"
    ),
    2202: "You do not have the required permissions in jira required to fetch boards for the project {}",
    2203: "ERROR: Failed downloading sprints for Jira board: {} with s_start_at={}.\nReceived 400 response:\n{}",
}


def generate_standard_error_msg(error_code, msg_args=[]):
    return f"[{error_code}] {ERROR_MESSAGES.get(error_code).format(*msg_args)}"


def log_standard_error(level, error_code, msg_args=[], exc_info=False):
    """
    For a failure that should be sent to the logger with an error_code, and also written
    to stdout (for user visibility)
    """
    assert level >= logging.WARNING
    msg = generate_standard_error_msg(error_code=error_code, msg_args=msg_args)
    logger.log(level=level, msg=msg)



import dataclasses
import gzip
import json
import logging
import os
import threading
import traceback
from pathlib import Path
from typing import Optional

import requests

from jf_ingest import logging_helper
from jf_ingest.config import IngestionConfig, IngestionType, JiraConfig
from jf_ingest.utils import ReauthSession, retry_session

logger = logging.getLogger(__name__)


class SubDirectory:
    JIRA = "jira"
    GIT = "git"


def _is_auth_error(exception_string: str) -> bool:
    """
    Checks if the exception string is an authentication error.
    :param exception_string: str
    :return: bool
    """
    for auth_str in [
        "authentication",
        "permission",
        "secreterror",
        "connecttimeouterror",
        "authorize",
    ]:
        if auth_str in exception_string:
            return True
    return False


class StrDefaultEncoder(json.JSONEncoder):
    def default(self, o):
        if dataclasses.is_dataclass(o):
            return dataclasses.asdict(o)
        return str(o)


class IngestIOHelper:
    def __init__(self, ingest_config: IngestionConfig):
        self.ingest_config = ingest_config
        # EVERYTHING in this file path will (potentially) be uploaded to S3!
        # DO NOT put any creds file in this path!!!!
        self.local_file_path = ingest_config.local_file_path

        if not os.path.exists(self.local_file_path):
            os.makedirs(self.local_file_path)

    def _get_file_name(self, object_name: str, batch_number: Optional[int] = 0):
        return f'{object_name}{batch_number if batch_number else ""}.json'

    def _write_file(self, json_data: dict | list[dict], full_file_path: str):
        """
        Writes json file data to local file system at filepath. Called by write_json_data_to_local.
        :param json_data:
        :param full_file_path:
        :return: nothing
        """
        with open(full_file_path, "wb") as f:
            f.write(json.dumps(json_data, indent=2, cls=StrDefaultEncoder).encode("utf-8"))
            logger.debug(f"File: {full_file_path}, Size: {round(f.tell() / 1000000, 1)}MB")

    def write_json_data_to_local(
        self,
        object_name: str,
        json_data: dict | list[dict],
        subdirectory: SubDirectory,
        batch_number: Optional[int] = 0,
    ) -> None:
        """
        Writes json data to local file system in proper subdirectory (depending on jira or git)
        :param object_name: from JiraObject or GitObject enum, eg "jira_boards", "git_prs", etc
        :param json_data: data returned from ingestion
        :param subdirectory: directory under `<output_dir>/<timestamp>/` to write to, eg "jira" or "git"
        :param batch_number: if this is a batched file, the batch number
        :return: None
        """
        try:
            file_name = self._get_file_name(object_name=object_name, batch_number=batch_number)
            full_file_path = f"{self.local_file_path}/{subdirectory}/{file_name}"
            logger.info(f"Attempting to save {object_name} data to {full_file_path}")
            try:
                self._write_file(json_data, full_file_path)
            except FileNotFoundError as e:
                Path(full_file_path).parent.mkdir(exist_ok=True, parents=True)
                self._write_file(json_data, full_file_path)
        except Exception as e:
            logger.error(
                f"Exception encountered when attempting to write data to local file! Error: {e}"
            )

    def get_signed_urls(
        self, filenames: list[str], timestamp: str, ingestion_type: IngestionType
    ) -> dict[str, dict]:
        """
        Gets signed urls from Jellyfish API for uploading to S3
        :param filenames: list of filenames that will need urls
        :param timestamp: timestamp created at beginning of ingestion (used for output directory)
        :return: the signed urls dict of shape {
            '<filename>': {
                's3_path': '<s3_path>',
                'url': {
                    'url': '<signed_url>',
                    'fields': {
                        '<field>': '<value>' for field in (key, AWSAccessKeyId, x-amz-security-token, policy, signature)
                    }
                }
            }
        }
        """
        base_url = self.ingest_config.jellyfish_api_base
        headers = {"Jellyfish-API-Token": self.ingest_config.jellyfish_api_token}
        payload = {
            "files": [f"{name}.gz" for name in filenames],
            "ingestType": ingestion_type,
        }  # the files will be gzipped before uploading

        r = requests.post(
            f"{base_url}/endpoints/ingest/signed-url?timestamp={timestamp}",
            headers=headers,
            json=payload,
        )
        r.raise_for_status()

        return r.json()["signed_urls"]

    def upload_file_to_s3(
        self, file_obj: bytes, s3_obj_path: str, signed_url: dict, dry_run=False
    ) -> None:
        """
        Uploads a given file object to the s3 path using the signed url
        :param file_obj: python file object opened in bytes mode
        :param s3_obj_path: s3 path; <company_slug>/<timestamp>/<filepath>
        :param signed_url: dict with signed url value and AWS fields
        :param dry_run: bool, set to True to prevent actual upload
        :return: None (will call raise_for_status on the upload response)
        """
        if dry_run:
            return
        session = retry_session()
        upload_resp = session.post(
            signed_url["url"],
            data=signed_url["fields"],
            files={"file": (s3_obj_path, file_obj)},
        )
        upload_resp.raise_for_status()
        return

    def wrapped_upload_file_to_s3(self, filepath: str, s3_obj_path: str, signed_url: dict) -> bool:
        """
        Wraps upload_file_to_s3 with error handling
        :param filepath: full filepath to local file
        :param s3_obj_path: s3 path; <company_slug>/<timestamp>/<filepath>
        :param signed_url: dict with signed url value and AWS fields
        :return: bool, True/False on Success/Failure
        """
        try:
            with open(filepath, "rb") as file_obj:
                logger.info(f"Compressing {filepath}...")
                gzip_object = gzip.compress(file_obj.read())
                self.upload_file_to_s3(gzip_object, s3_obj_path, signed_url)
                logger.info(f"Successfully uploaded {filepath} to S3")

                return True
        except Exception as e:
            if _is_auth_error(str(e).lower()):
                logger.error(
                    f"Authentication error encountered when uploading {filepath} to S3! Error: \n{e} "
                    f"\nIs there a proxy or firewall blocking the connection?"
                )
            else:
                logger.error(f"Exception encountered when uploading {filepath} to S3! Error: {e}")
            logger.debug(traceback.format_exc())

            return False

    def upload_local_files_to_s3_threaded(self, filenames: list[str], prefix: str = "") -> bool:
        """
        Uploads a list of local files to S3 using a thread pool.
        :param filenames: list[str]
        :param prefix: str
        :return: bool, True/False on Success/Failure
        """
        filenames_for_s3 = [f"{prefix}{Path(filename).name}" for filename in filenames]
        signed_urls = self.get_signed_urls(
            filenames_for_s3,
            self.ingest_config.timestamp,
            ingestion_type=self.ingest_config.ingest_type,
        )
        threads = [
            threading.Thread(
                target=self.wrapped_upload_file_to_s3,
                args=[
                    f"{self.local_file_path}/{filename.replace('.gz', '')}",
                    url_dict["s3_path"],
                    url_dict["url"],
                ],
            )
            for (filename, url_dict) in signed_urls.items()
        ]

        logger.info(f"Uploading {len(filenames)} files to S3 using {len(threads)} threads...")

        thread_exceptions = []

        for t in threads:
            t.start()

        for t in threads:
            t.join()

        success = True

        if any(thread_exceptions):
            # Run through exceptions and inject them into the agent log
            for exception in thread_exceptions:
                logger.error(
                    f"A thread exception was encountered when uploading files to S3: {exception}"
                )
            logger.error(
                "There was an error uploading files to S3. Please check the agent log for more details. When "
                "connectivity has been restored you may run the agent in `send-only` mode to retry sending the files."
            )
            success = False
            logger.info("✗")
        if success:
            logger.info("✓")

        return success

    def upload_all_directory_output_to_s3(self, directory: str) -> bool:
        """
        Uploads all files in a given directory to S3
        :param directory: string directory name
        :return: bool, True/False on Success/Failure
        """
        logger.info(f"Uploading all {directory} output to S3...")
        success = False
        try:
            _, directories, filenames = next(os.walk(f"{self.local_file_path}/{directory}/"))
            filenames = [f"{self.local_file_path}/{directory}/{filename}" for filename in filenames]
            success = self.upload_local_files_to_s3_threaded(filenames, prefix=f"{directory}/")
        except StopIteration:
            logger.warning(f"No {directory} output to upload, skipping.")
        if success:
            logger.info(f"Successfully compressed and uploaded {directory} output to S3")

        return success

    def upload_all_locals_to_s3(self) -> None:
        """
        Uploads all local files to S3
        :return: None
        """
        success_jira = self.upload_all_directory_output_to_s3(SubDirectory.JIRA)
        success_git = self.upload_all_directory_output_to_s3(SubDirectory.GIT)

        logger.info("Uploading all remaining files to S3...")
        success_remaining = self.upload_all_directory_output_to_s3("./")

        logger.info("Uploading agent log to S3...")
        success_log = self.upload_local_files_to_s3_threaded([logging_helper.LOG_FILE_NAME])
        if success_log:
            logger.info("✓")
        else:
            logger.info("✗")

        if success_jira and success_git and success_remaining:
            done_file_dict = self.get_signed_urls(
                [".done"],
                timestamp=self.ingest_config.timestamp,
                ingestion_type=self.ingest_config.ingest_type,
            )[".done"]
            self.upload_file_to_s3(b".done", done_file_dict["s3_path"], done_file_dict["url"])
